unit XmlIni;

interface

{ 
  E ]p
  
  E-Mail: howardchen@pie.com.tw
}

uses
  Classes, XMLIntf;

type
  TXmlIni = class;
  TXmlIniSection = class;

  TXmlIniKey = class(TObject)
  private
    FOwner: TXmlIniSection;
    FText: string;
    FName: string;
    FStream: TMemoryStream;
    function GetAsInteger: Integer;
    procedure SetAsInteger(const Value: Integer);
    function GetAsBool: Boolean;
    procedure SetAsBool(const Value: Boolean);
    function GetAsStream: TMemoryStream;
    procedure SetAsStream(const Value: TMemoryStream);
    function GetText: string;
    procedure SetText(const Value: string);
  public
    constructor Create(AOwner: TXmlIniSection; AName: string; AText: string = '');
    destructor Destroy; override;

    procedure Clear;

    property Name: string read FName;
    property Text: string read GetText write SetText;
    property AsInteger: Integer read GetAsInteger write SetAsInteger;
    property AsBool: Boolean read GetAsBool write SetAsBool;
    property AsStream: TMemoryStream read GetAsStream write SetAsStream;
  end;

  TXmlIniSection = class(TObject)
  private
    FOwner: TXmlIni;
    FList: TList;
    FName: string;
    function InternalFind(KeyName: string): TXmlIniKey;
    function InternalAdd(KeyName: string): TXmlIniKey;
    function GetCount: Integer;
    function GetKeyByIndex(Index: Integer): TXmlIniKey;
    function GetKey(Index: string): TXmlIniKey;
    function GetItem(Index: string): string;
    procedure SetItem(Index: string; const Value: string);
  public
    constructor Create(AOwner: TXmlIni; AName: string);
    destructor Destroy; override;

    property Name: string read FName;
    property KeyByIndex[Index: Integer]: TXmlIniKey read GetKeyByIndex;
    property Key[Index: string]: TXmlIniKey read GetKey;
    property Item[Index: string]: string read GetItem write SetItem; default;
    property Count: Integer read GetCount;

    procedure Clear;
    procedure Collation;
    function Exists(KeyName: string): Boolean;
  end;

  TXmlIni = class(TObject)
  private
    FList: TList;
    FFileName: string;
    FAutoSave: Boolean;
    FIgnoreCase: Boolean;
    FCompressed: Boolean;
    procedure InternalLoadFromFile(AFileName: string);
    procedure InternalSaveToFile(AFileName: string);
    function InternalFind(SectionName: string): TXmlIniSection;
    function InternalAdd(SectionName: string): TXmlIniSection;
    function GetCount: Integer;
    function GetSectionByIndex(Index: Integer): TXmlIniSection;
    function GetSectionByName(Index: string): TXmlIniSection;
    procedure SetFileName(const Value: string);
    procedure SetAutoSave(const Value: Boolean);
    procedure SetCompressed(const Value: Boolean);
  public
    constructor Create(AFileName: string = '');
    destructor Destroy; override;

    property FileName: string read FFileName write SetFileName;
    property AutoSave: Boolean read FAutoSave write SetAutoSave;

    procedure LoadFromFile(FileName: string = '');
    procedure SaveToFile(FileName: string = '');

    procedure LoadFromStream(Stream: TStream; EncodingType: TXMLEncodingType = xetUnknown);
    procedure SaveToStream(Stream: TStream);

    procedure LoadFromString(Value: string);
    function SaveToString: string;

    procedure LoadFromBase64String(Value: string);
    function SaveToBase64String: string;

    procedure Clear;
    procedure Collation;

    property IgnoreCase: Boolean read FIgnoreCase write FIgnoreCase;
    property Compressed: Boolean read FCompressed write SetCompressed;

    property SectionByIndex[Index: Integer]: TXmlIniSection read GetSectionByIndex;
    property Section[Index: string]: TXmlIniSection read GetSectionByName; default;
    property Count: Integer read GetCount;

    function Add(SectionName: string): TXmlIniSection;
    procedure Delete(SectionName: string);
    function Exists(SectionName: string; KeyName: string = ''): Boolean;
  end;

implementation

uses
  Forms, xmldom, msxmldom, XMLDoc, zlib, SysUtils;

const
  XmlTransStrSrc: array[0..1] of string = (#13, #10);
  XmlTransStrDest: array[0..1] of string = ('&#xD;', '&#xA;');

function TransStrToXml(Str: string): string;
var
  i: Integer;
begin
  Result := Str;
  for i := Low(XmlTransStrSrc) to High(XmlTransStrSrc) do
    Result := StringReplace(Result, XmlTransStrSrc[i], XmlTransStrDest[i], [rfReplaceAll, rfIgnoreCase]);
end;

function TransXmlToStr(Xml: string): string;
var
  i: Integer;
begin
  Result := Xml;
  for i := Low(XmlTransStrSrc) to High(XmlTransStrSrc) do
    Result := StringReplace(Result, XmlTransStrDest[i], XmlTransStrSrc[i], [rfReplaceAll, rfIgnoreCase]);
end;

const
  BufSize = 1024;

procedure CompressStream(SourceStream, DestStream: TStream);
var
  CS: TCompressionStream;
  Buf: PChar;
  DataSize, ReadCount: LongInt;
begin
  SourceStream.Position := 0;
  DestStream.Size := 0;

  DataSize := SourceStream.Size;
  DestStream.Write('ZL', 2);
  DestStream.Write(DataSize, SizeOf(LongInt));

  CS := TCompressionStream.Create(clMax, DestStream);
  try
    Buf := AllocMem(BufSize);
    try
      repeat
        ReadCount := SourceStream.Read(Buf^, BufSize);
        if ReadCount > 0 then
          CS.Write(Buf^, ReadCount);
      until ReadCount = 0;
    finally
      FreeMem(Buf);
    end;
  finally
    CS.Free;
  end;
end;

procedure DecompressStream(SourceStream, DestStream: TStream);
var
  DS: TDecompressionStream;
  Ident: array[0..1] of Char;
  Buf: PChar;
  DataSize, ReadCount: LongInt;
begin
  SourceStream.Position := 0;
  SourceStream.Read(Ident, 2);
  SourceStream.Read(DataSize, SizeOf(LongInt));
  if (Ident = 'ZL') and (DataSize > 0) then
  begin
    DestStream.Size := 0;
    DS := TDecompressionStream.Create(SourceStream);
    try
      Buf := AllocMem(BufSize);
      try
        while DataSize > 0 do
        begin
          ReadCount := DataSize;
          if ReadCount > BufSize then
            ReadCount := BufSize;
          DS.Read(Buf^, ReadCount);
          DestStream.Write(Buf^, ReadCount);
          Dec(DataSize, ReadCount);
        end;
      finally
        FreeMem(Buf);
      end;
    finally
      DS.Free;
    end;
  end;
end;

type
  PByte4 = ^TByte4;
  TByte4 = packed record
    b1: Byte;
    b2: Byte;
    b3: Byte;
    b4: Byte;
  end;

  PByte3 = ^TByte3;
  TByte3 = packed record
    b1: Byte;
    b2: Byte;
    b3: Byte;
  end;

const
  BUFFER_SIZE = $3000;
 { Caution: For MimeEncodeStream and all other kinds of multi-buffered
   Mime encodings (i.e. Files etc.), BufferSize must be set to a multiple of 3.
   Even though the implementation of the Mime decoding routines below
   does not require a particular buffer size, they work fastest with sizes of
   multiples of four. The chosen size is a multiple of 3 and of 4 as well.
   The following numbers are, in addition, also divisible by 1024:
   $2400, $3000, $3C00, $4800, $5400, $6000, $6C00. }

  MIME_ENCODE_TABLE: array[0..63] of Byte = (
    065, 066, 067, 068, 069, 070, 071, 072, // 00 - 07
    073, 074, 075, 076, 077, 078, 079, 080, // 08 - 15
    081, 082, 083, 084, 085, 086, 087, 088, // 16 - 23
    089, 090, 097, 098, 099, 100, 101, 102, // 24 - 31
    103, 104, 105, 106, 107, 108, 109, 110, // 32 - 39
    111, 112, 113, 114, 115, 116, 117, 118, // 40 - 47
    119, 120, 121, 122, 048, 049, 050, 051, // 48 - 55
    052, 053, 054, 055, 056, 057, 043, 047); // 56 - 63

  MIME_DECODE_TABLE: array[Byte] of Cardinal = (
    255, 255, 255, 255, 255, 255, 255, 255, //  00 -  07
    255, 255, 255, 255, 255, 255, 255, 255, //  08 -  15
    255, 255, 255, 255, 255, 255, 255, 255, //  16 -  23
    255, 255, 255, 255, 255, 255, 255, 255, //  24 -  31
    255, 255, 255, 255, 255, 255, 255, 255, //  32 -  39
    255, 255, 255, 062, 255, 255, 255, 063, //  40 -  47
    052, 053, 054, 055, 056, 057, 058, 059, //  48 -  55
    060, 061, 255, 255, 255, 255, 255, 255, //  56 -  63
    255, 000, 001, 002, 003, 004, 005, 006, //  64 -  71
    007, 008, 009, 010, 011, 012, 013, 014, //  72 -  79
    015, 016, 017, 018, 019, 020, 021, 022, //  80 -  87
    023, 024, 025, 255, 255, 255, 255, 255, //  88 -  95
    255, 026, 027, 028, 029, 030, 031, 032, //  96 - 103
    033, 034, 035, 036, 037, 038, 039, 040, // 104 - 111
    041, 042, 043, 044, 045, 046, 047, 048, // 112 - 119
    049, 050, 051, 255, 255, 255, 255, 255, // 120 - 127
    255, 255, 255, 255, 255, 255, 255, 255,
    255, 255, 255, 255, 255, 255, 255, 255,
    255, 255, 255, 255, 255, 255, 255, 255,
    255, 255, 255, 255, 255, 255, 255, 255,
    255, 255, 255, 255, 255, 255, 255, 255,
    255, 255, 255, 255, 255, 255, 255, 255,
    255, 255, 255, 255, 255, 255, 255, 255,
    255, 255, 255, 255, 255, 255, 255, 255,
    255, 255, 255, 255, 255, 255, 255, 255,
    255, 255, 255, 255, 255, 255, 255, 255,
    255, 255, 255, 255, 255, 255, 255, 255,
    255, 255, 255, 255, 255, 255, 255, 255,
    255, 255, 255, 255, 255, 255, 255, 255,
    255, 255, 255, 255, 255, 255, 255, 255,
    255, 255, 255, 255, 255, 255, 255, 255,
    255, 255, 255, 255, 255, 255, 255, 255);

function MimeEncodedSize(const i: Integer): Integer;
begin
  Result := (i + 2) div 3 * 4;
end;

procedure MimeEncode(const InputBuffer: Pointer; const InputByteCount: Integer; const OutputBuffer: Pointer);
var
  b: Cardinal;
  InMax3: Integer;
  pIn, PInLimit: ^Byte;
  POut: PByte4;
begin
{$IFDEF RangeChecking}
  if InputBuffer = nil then raise EMime.Create(SInputBufferNil);
  if OutputBuffer = nil then raise EMime.Create(SOutputBufferNil);
{$ENDIF}
  if InputByteCount <= 0 then Exit;

  pIn := InputBuffer;
  InMax3 := InputByteCount div 3 * 3;
  POut := OutputBuffer;

  Integer(PInLimit) := Integer(pIn) + InMax3;

  while pIn <> PInLimit do
  begin
    b := pIn^;
    b := b shl 8;
    Inc(pIn);
    b := b or pIn^;
    b := b shl 8;
    Inc(pIn);
    b := b or pIn^;
    Inc(pIn);

   // Write 4 bytes to OutputBuffer (in reverse order).
    POut.b4 := MIME_ENCODE_TABLE[b and $3F];
    b := b shr 6;
    POut.b3 := MIME_ENCODE_TABLE[b and $3F];
    b := b shr 6;
    POut.b2 := MIME_ENCODE_TABLE[b and $3F];
    b := b shr 6;
    POut.b1 := MIME_ENCODE_TABLE[b];

    Inc(POut);
  end;

  case InputByteCount - InMax3 of
    1:
      begin
        b := pIn^;

        b := b shl 4;

        POut.b2 := MIME_ENCODE_TABLE[b and $3F];
        b := b shr 6;
        POut.b1 := MIME_ENCODE_TABLE[b];

        POut.b3 := Ord('='); // Fill remaining 2 bytes.
        POut.b4 := Ord('=');
      end;
    2:
      begin
        b := pIn^;
        Inc(pIn);
        b := b shl 8;
        b := b or pIn^;

        b := b shl 2;

        POut.b3 := MIME_ENCODE_TABLE[b and $3F];
        b := b shr 6;
        POut.b2 := MIME_ENCODE_TABLE[b and $3F];
        b := b shr 6;
        POut.b1 := MIME_ENCODE_TABLE[b];

        POut.b4 := Ord('='); // Fill remaining byte.
      end;
  end;
end;

procedure MimeEncodeStream(const InputStream: TStream; const OutputStream: TStream);
var
  InputBuffer: array [0..BUFFER_SIZE - 1] of Byte;
  OutputBuffer: array [0..((BUFFER_SIZE + 2) div 3) * 4 - 1] of Byte;
  BytesRead: Integer;
begin
  BytesRead := InputStream.Read(InputBuffer, SizeOf(InputBuffer));
  while BytesRead = SizeOf(InputBuffer) do
  begin
    MimeEncode(@InputBuffer, SizeOf(InputBuffer), @OutputBuffer);
    OutputStream.Write(OutputBuffer, SizeOf(OutputBuffer));
    BytesRead := InputStream.Read(InputBuffer, SizeOf(InputBuffer));
  end;
  if BytesRead > 0 then
  begin
    MimeEncode(@InputBuffer, BytesRead, @OutputBuffer);
    OutputStream.Write(OutputBuffer, MimeEncodedSize(BytesRead));
  end;
end;

function MimeDecodedSize(const i: Integer): Integer;
begin
  Result := (i + 3) div 4 * 3;
end;

function MimeDecodePartialEnd(const OutputBuffer: Pointer; const ByteBuffer: Cardinal; const ByteBufferSpace: Cardinal): Integer;
var
  lByteBuffer: Cardinal;
begin
{$IFDEF RangeChecking}
  if OutputBuffer = nil then raise EMime.Create(SOutputBufferNil);
{$ENDIF}
  case ByteBufferSpace of
    1:
      begin
        lByteBuffer := ByteBuffer shr 2;
        PByte3(OutputBuffer).b2 := Byte(lByteBuffer);
        lByteBuffer := lByteBuffer shr 8;
        PByte3(OutputBuffer).b1 := Byte(lByteBuffer);
        Result := 2;
      end;
    2:
      begin
        lByteBuffer := ByteBuffer shr 4;
        PByte3(OutputBuffer).b1 := Byte(lByteBuffer);
        Result := 1;
      end;
  else
    Result := 0;
  end;
end;

function MimeDecodePartial(const InputBuffer: Pointer; const InputBytesCount: Integer; const OutputBuffer: Pointer; var ByteBuffer: Cardinal; var ByteBufferSpace: Cardinal): Integer;
var
  lByteBuffer, lByteBufferSpace, c: Cardinal;
  pIn, PInLimit: ^Byte;
  POut: PByte3;
begin
{$IFDEF RangeChecking}
  if InputBuffer = nil then raise EMime.Create(SInputBufferNil);
  if OutputBuffer = nil then raise EMime.Create(SOutputBufferNil);
{$ENDIF}
  if InputBytesCount > 0 then
  begin
    pIn := InputBuffer;
    Integer(PInLimit) := Integer(pIn) + InputBytesCount;

    POut := OutputBuffer;

    lByteBuffer := ByteBuffer;
    lByteBufferSpace := ByteBufferSpace;

    while pIn <> PInLimit do
    begin
      c := MIME_DECODE_TABLE[pIn^]; // Read from InputBuffer.
      Inc(pIn);

      if c = $FF then Continue;

      lByteBuffer := lByteBuffer shl 6;
      lByteBuffer := lByteBuffer or c;
      Dec(lByteBufferSpace);

      if lByteBufferSpace <> 0 then Continue; // Read 4 bytes from InputBuffer?

      POut.b3 := Byte(lByteBuffer); // Write 3 bytes to OutputBuffer (in reverse order).
      lByteBuffer := lByteBuffer shr 8;
      POut.b2 := Byte(lByteBuffer);
      lByteBuffer := lByteBuffer shr 8;
      POut.b1 := Byte(lByteBuffer);
      lByteBuffer := 0;
      Inc(POut);

      lByteBufferSpace := 4;
    end;

    ByteBuffer := lByteBuffer;
    ByteBufferSpace := lByteBufferSpace;

    Result := Integer(POut) - Integer(OutputBuffer);
  end
  else
    Result := 0;
end;

procedure MimeDecodeStream(const InputStream: TStream; const OutputStream: TStream);
var
  ByteBuffer, ByteBufferSpace: Cardinal;
  InputBuffer: array [0..BUFFER_SIZE - 1] of Byte;
  OutputBuffer: array [0..(BUFFER_SIZE + 3) div 4 * 3 - 1] of Byte;
  BytesRead: Integer;
begin
  ByteBuffer := 0;
  ByteBufferSpace := 4;
  BytesRead := InputStream.Read(InputBuffer, SizeOf(InputBuffer));
  while BytesRead > 0 do
  begin
    OutputStream.Write(OutputBuffer, MimeDecodePartial(@InputBuffer, BytesRead, @OutputBuffer, ByteBuffer, ByteBufferSpace));
    BytesRead := InputStream.Read(InputBuffer, SizeOf(InputBuffer));
  end;
  OutputStream.Write(OutputBuffer, MimeDecodePartialEnd(@OutputBuffer, ByteBuffer, ByteBufferSpace));
end;

{ TXmlIni }

constructor TXmlIni.Create(AFileName: string);
begin
  FList := TList.Create;
  FAutoSave := True;
  FFileName := '';
  FIgnoreCase := True;
  FCompressed := False;

  LoadFromFile(AFileName);
end;

destructor TXmlIni.Destroy;
begin
  if FAutoSave then
    SaveToFile;
  Clear;
  FList.Free;
end;

procedure TXmlIni.LoadFromFile(FileName: string = '');
begin
  if FileName <> '' then
    FFileName := FileName;
  InternalLoadFromFile(FFileName);
end;

procedure TXmlIni.SaveToFile(FileName: string = '');
begin
  if FileName <> '' then
    FFileName := FileName;
  InternalSaveToFile(FFileName);
end;

procedure TXmlIni.Clear;
var
  i: Integer;
begin
  for i := Count - 1 downto 0 do
  begin
    SectionByIndex[i].Free;
    FList.Delete(i);
  end;
end;

procedure TXmlIni.Collation;
var
  i: Integer;
begin
  for i := Count - 1 downto 0 do
  begin
    SectionByIndex[i].Collation;
    if SectionByIndex[i].Count = 0 then
    begin
      SectionByIndex[i].Free;
      FList.Delete(i);
    end;
  end;
end;

function TXmlIni.Add(SectionName: string): TXmlIniSection;
begin
  Result := InternalFind(SectionName);
  if Result = nil then
    Result := InternalAdd(SectionName);
end;

procedure TXmlIni.Delete(SectionName: string);
var
  i: Integer;
begin
  for i := Count - 1 downto 0 do
    if SectionByIndex[i].Name = SectionName then
    begin
      SectionByIndex[i].Free;
      FList.Delete(i);
      Break;
    end;
end;

function TXmlIni.Exists(SectionName: string; KeyName: string = ''): Boolean;
var
  Section: TXmlIniSection;
begin
  Result := False;

  Section := InternalFind(SectionName);
  if Section <> nil then
  begin
    if KeyName <> '' then
      Result := Section.Exists(KeyName)
    else
      Result := True;
  end;
end;

procedure TXmlIni.LoadFromStream(Stream: TStream; EncodingType: TXMLEncodingType = xetUnknown);
var
  XML: TXMLDocument;
  Root: IXmlNode;
  i: Integer;
  j: Integer;
  TestBuf: array[0..1] of Char;
  UnzipMS: TMemoryStream;
begin
  Clear;

  try
    XML := TXMLDocument.Create(Application);
    try
      Stream.Position := 0;
      Stream.Read(TestBuf, 2);

      if TestBuf = 'ZL' then
      begin
        FCompressed := True;
        Stream.Position := 0;
        UnzipMS := TMemoryStream.Create;
        try
          DecompressStream(Stream, UnzipMS);
          XML.LoadFromStream(UnzipMS, EncodingType);
        finally
          UnzipMS.Free;
        end;
      end
      else
      begin
        FCompressed := False;
        Stream.Position := 0;
        XML.LoadFromStream(Stream, EncodingType);
      end;

      Root := XML.DocumentElement;

      if Root.NodeName <> 'xmlini' then
        Exit;

      for i := 0 to Root.ChildNodes.Count - 1 do
        if Root.ChildNodes[i].NodeName = 'section' then
          with Add(TransXmlToStr(Root.ChildNodes[i].Attributes['name'])) do
          begin
            for j := 0 to Root.ChildNodes[i].ChildNodes.Count - 1 do
              if Root.ChildNodes[i].ChildNodes[j].NodeName = 'key' then
                Item[TransXmlToStr(Root.ChildNodes[i].ChildNodes[j].Attributes['name'])] :=
                  TransXmlToStr(Root.ChildNodes[i].ChildNodes[j].Text);
          end;

      Collation;
    finally
      XML.Free;
    end;
  except
  end;
end;

procedure TXmlIni.SaveToStream(Stream: TStream);
var
  XML: TXMLDocument;
  Root, SectionNode, KeyNode: IXmlNode;
  i: Integer;
  j: Integer;
  UnzipMS: TMemoryStream;
begin
  Collation;

  XML := TXMLDocument.Create(Application);
  try
    XML.Active := True;
    XML.Options := XML.Options + [doNodeAutoIndent];
    XML.NodeIndentStr := '  ';

    Root := XML.AddChild('xmlini');

    for i := 0 to Count - 1 do
    begin
      SectionNode := Root.AddChild('section');
      SectionNode.Attributes['name'] := TransStrToXml(SectionByIndex[i].Name);
      for j := 0 to SectionByIndex[i].Count - 1 do
      begin
        KeyNode := SectionNode.AddChild('key');
        KeyNode.Attributes['name'] := TransStrToXml(SectionByIndex[i].KeyByIndex[j].Name);
        KeyNode.Text := TransStrToXml(SectionByIndex[i].KeyByIndex[j].Text);
      end;
    end;

    XML.StandAlone := 'yes';
    XML.Version := '1.0';
    XML.Encoding := 'UTF-16';

    if FCompressed then
    begin
      UnzipMS := TMemoryStream.Create;
      try
        XML.SaveToStream(UnzipMS);
        UnzipMS.Position := 0;
        Stream.Position := 0;
        CompressStream(UnzipMS, Stream);
      finally
        UnzipMS.Free;
      end;
    end
    else
    begin
      Stream.Position := 0;
      if Stream is TStringStream then
        TStringStream(Stream).WriteString(XML.XML.Text)
      else
        XML.SaveToStream(Stream);
    end;
  finally
    XML.Free;
  end;
end;

procedure TXmlIni.LoadFromString(Value: string);
var
  SS: TStringStream;
begin
  if TrimLeft(Value) = '' then
    Clear
  else if TrimLeft(Value)[1] = '<' then
  begin
    SS := TStringStream.Create(Value);
    try
      LoadFromStream(SS);
    finally
      SS.Free;
    end;
  end
  else
    LoadFromBase64String(Value);
end;

function TXmlIni.SaveToString: string;
var
  SS: TStringStream;
begin
  if not FCompressed then
  begin
    SS := TStringStream.Create('');
    try
      SaveToStream(SS);
      Result := SS.DataString;
    finally
      SS.Free;
    end;
  end
  else
    Result := SaveToBase64String;
end;

procedure TXmlIni.LoadFromBase64String(Value: string);
var
  SS: TStringStream;
  MS: TMemoryStream;
begin
  if (TrimLeft(Value) = '') or (TrimLeft(Value)[1] = '<') then
    LoadFromString(Value)
  else
  begin
    SS := TStringStream.Create(Trim(Value));
    try
      MS := TMemoryStream.Create;
      try
        SS.Position := 0;
        MimeDecodeStream(SS, MS);
        LoadFromStream(MS);
      finally
        MS.Free;
      end;
    finally
      SS.Free;
    end;
  end;
end;

function TXmlIni.SaveToBase64String: string;
var
  SS: TStringStream;
  MS: TMemoryStream;
begin
  MS := TMemoryStream.Create;
  try
    SaveToStream(MS);

    SS := TStringStream.Create('');
    try
      MS.Position := 0;
      MimeEncodeStream(MS, SS);
      Result := SS.DataString;
    finally
      SS.Free;
    end;
  finally
    MS.Free;
  end;
end;

procedure TXmlIni.InternalLoadFromFile(AFileName: string);
var
  MS: TMemoryStream;
begin
  Clear;

  if not FileExists(aFileName) then
    Exit;

  MS := TMemoryStream.Create;
  try
    MS.LoadFromFile(aFileName);
    LoadFromStream(MS);
  except
    MS.Free;
  end;
end;

procedure TXmlIni.InternalSaveToFile(AFileName: string);
var
  MS: TMemoryStream;
begin
  if AFileName = '' then
    Exit;

  MS := TMemoryStream.Create;
  try
    SaveToStream(MS);
    MS.SaveToFile(AFileName);
  except
    MS.Free;
  end;
end;

function TXmlIni.InternalFind(SectionName: string): TXmlIniSection;
var
  i: Integer;
begin
  Result := nil;

  for i := Count - 1 downto 0 do
    if (SectionByIndex[i].Name = SectionName) or
      (FIgnoreCase and (UpperCase(SectionByIndex[i].Name) = UpperCase(SectionName))) then
    begin
      Result := SectionByIndex[i];
      Exit;
    end;
end;

function TXmlIni.InternalAdd(SectionName: string): TXmlIniSection;
begin
  Result := TXmlIniSection.Create(Self, SectionName);
  FList.Add(Result);
end;

function TXmlIni.GetCount: Integer;
begin
  Result := FList.Count;
end;

function TXmlIni.GetSectionByIndex(Index: Integer): TXmlIniSection;
begin
  Result := TXmlIniSection(FList[Index]);
end;

function TXmlIni.GetSectionByName(Index: string): TXmlIniSection;
begin
  Result := Add(Index);
end;

procedure TXmlIni.SetFileName(const Value: string);
begin
  if FFileName <> Value then
  begin
    if FAutoSave then
      SaveToFile;
    FFileName := Value;
    LoadFromFile;
  end;
end;

procedure TXmlIni.SetAutoSave(const Value: Boolean);
begin
  FAutoSave := Value;
  if FAutoSave then
    SaveToFile;
end;

procedure TXmlIni.SetCompressed(const Value: Boolean);
begin
  if FCompressed <> Value then
  begin
    FCompressed := Value;
    if FAutoSave then
      SaveToFile;
  end;
end;

{ TXmlIniSection }

constructor TXmlIniSection.Create(AOwner: TXmlIni; AName: string);
begin
  FOwner := AOwner;

  FList := TList.Create;
  FName := AName;
end;

destructor TXmlIniSection.Destroy;
begin
  Clear;
  FList.Free;

  inherited;
end;

procedure TXmlIniSection.Clear;
var
  i: Integer;
begin
  for i := Count - 1 downto 0 do
  begin
    KeyByIndex[i].Free;
    FList.Delete(i);
  end;
end;

procedure TXmlIniSection.Collation;
var
  i: Integer;
begin
  for i := Count - 1 downto 0 do
    if KeyByIndex[i].Text = '' then
    begin
      KeyByIndex[i].Free;
      FList.Delete(i);
    end;
end;

function TXmlIniSection.Exists(KeyName: string): Boolean;
begin
  Result := InternalFind(KeyName) <> nil;
end;

function TXmlIniSection.GetCount: Integer;
begin
  Result := FList.Count;
end;

function TXmlIniSection.GetKeyByIndex(Index: Integer): TXmlIniKey;
begin
  Result := TXmlIniKey(FList[Index]);
end;

function TXmlIniSection.GetKey(Index: string): TXmlIniKey;
var
  Key: TXmlIniKey;
begin
  Key := InternalFind(Index);
  if Key = nil then
    Key := InternalAdd(Index);
  Result := Key;
end;

function TXmlIniSection.InternalAdd(KeyName: string): TXmlIniKey;
begin
  Result := TXmlIniKey.Create(Self, KeyName);
  FList.Add(Result);
end;

function TXmlIniSection.InternalFind(KeyName: string): TXmlIniKey;
var
  i: Integer;
begin
  Result := nil;

  for i := 0 to Count - 1 do
    if (KeyByIndex[i].Name = KeyName) or
      (FOwner.FIgnoreCase and (UpperCase(KeyByIndex[i].Name) = UpperCase(KeyName))) then
    begin
      Result := KeyByIndex[i];
      Break;
    end;
end;

function TXmlIniSection.GetItem(Index: string): string;
begin
  Result := Key[Index].Text;
end;

procedure TXmlIniSection.SetItem(Index: string; const Value: string);
begin
  Key[Index].Text := Value;
end;

{ TXmlIniKey }

constructor TXmlIniKey.Create(AOwner: TXmlIniSection; AName: string; AText: string = '');
begin
  FOwner := AOwner;

  FName := AName;
  FText := AText;
  FStream := nil;
end;

destructor TXmlIniKey.Destroy;
begin
  if FStream <> nil then
    FStream.Free;

  inherited;
end;

function TXmlIniKey.GetAsStream: TMemoryStream;
var
  SS: TStringStream;
  Bas64Text: string;
begin
  if FStream = nil then
  begin
    FStream := TMemoryStream.Create;

    Bas64Text := Trim(FText);
    if (Bas64Text <> '') and (Copy(Bas64Text, 1, 8) = '?BASE64?') then
    begin
      Delete(Bas64Text, 1, 8);
      try
        SS := TStringStream.Create(Bas64Text);
        try
          MimeDecodeStream(SS, FStream);
        finally
          SS.Free;
        end;
      except
      end;
    end;
  end;

  Result := FStream;
end;

procedure TXmlIniKey.SetAsStream(const Value: TMemoryStream);
begin
  if Value.Size = 0 then
    Clear
  else
  begin
    Value.Position := 0;
    if FStream = nil then
      FStream := TMemoryStream.Create;
    FStream.Clear;
    FStream.CopyFrom(Value, 0);
  end;
end;

procedure TXmlIniKey.Clear;
begin
  FText := '';
  if FStream <> nil then
  begin
    FStream.Free;
    FStream := nil;
  end;
end;

function TXmlIniKey.GetText: string;
var
  SS: TStringStream;
begin
  if (FStream <> nil) and (FStream.Size > 0) then
  begin
    SS := TStringStream.Create('');
    try
      FStream.Position := 0;
      MimeEncodeStream(FStream, SS);
      FText := '?BASE64?' + SS.DataString;
    finally
      SS.Free;
    end;
  end;
  Result := FText;
end;

procedure TXmlIniKey.SetText(const Value: string);
begin
  if FStream <> nil then
  begin
    FStream.Clear;
    FStream := nil;
  end;
  FText := Value;
end;

function TXmlIniKey.GetAsInteger: Integer;
begin
  Result := StrToIntDef(Text, 0);
end;

procedure TXmlIniKey.SetAsInteger(const Value: Integer);
begin
  Text := IntToStr(Value);
end;

function TXmlIniKey.GetAsBool: Boolean;
begin
  if (UpperCase(Text) = 'TRUE') or (UpperCase(Text) = 'YES') or
    (UpperCase(Text) = 'Y') or (UpperCase(Text) = '1') or (UpperCase(Text) = 'ON') then
    Result := True
  else
    Result := False;
end;

procedure TXmlIniKey.SetAsBool(const Value: Boolean);
begin
  if Value then
    Text := 'True'
  else
    Text := 'False';
end;

end.

